import gc
import hashlib
import json
import logging
import os
import sys
import threading
import urllib.parse
import urllib.request
import warnings
from os.path import expanduser
from pathlib import Path
from types import SimpleNamespace
from typing import Dict, Any, List, Tuple
from urllib.request import Request, urlopen

import pkg_resources
import regex as re
import torch
import yaml
from clip.simple_tokenizer import SimpleTokenizer, whitespace_clean, basic_clean
from packaging.version import Version
from spellchecker import SpellChecker
from tqdm.auto import tqdm

cache_dir = os.environ.get(
    'DISCOART_CACHE_DIR', os.path.join(expanduser('~'), '.cache', __package__)
)

from yaml import Loader

from . import __resources_path__


def _get_logger():
    logger = logging.getLogger(__package__)
    _log_level = os.environ.get('DISCOART_LOG_LEVEL', 'INFO')
    logger.setLevel(_log_level)
    ch = logging.StreamHandler()
    ch.setLevel(_log_level)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(ch)
    logger.propagate = False
    return logger


logger = _get_logger()


def get_model_list():
    with open(
        os.environ.get('DISCOART_MODELS_YAML', f'{__resources_path__}/models.yml')
    ) as ymlfile:
        return yaml.load(ymlfile, Loader=Loader)


models_list = get_model_list()


def get_remote_model_list(local_model_list: Dict[str, Any], force_print: bool = False):
    if 'DISCOART_DISABLE_REMOTE_MODELS' in os.environ:
        return

    try:
        req = Request(
            os.environ.get(
                'DISCOART_REMOTE_MODELS_URL',
                'https://raw.githubusercontent.com/jina-ai/discoart/main/discoart/resources/models.yml',
            ),
            headers={'User-Agent': 'Mozilla/5.0'},
        )
        with urlopen(
            req, timeout=2
        ) as resp:  # 'with' is important to close the resource after use
            remote_model_list = yaml.load(resp, Loader=Loader)
    except Exception as ex:
        logger.error(f'can not fetch the latest `model_list` from remote, {ex}')

    if (remote_model_list and remote_model_list != local_model_list) or force_print:
        if not force_print:
            logger.warning(
                'remote model list is different from the local model list, override local list.'
            )

        from rich.table import Table
        from rich import box, print

        param_str = Table(
            box=box.ROUNDED,
            highlight=True,
            title_justify='center',
        )
        param_str.add_column('Diffusion Model', justify='right')
        param_str.add_column('Remote', justify='left')
        param_str.add_column('Local', justify='left')
        param_str.add_column('Synced', justify='left')
        all_models = sorted(
            set(list(remote_model_list.keys()) + list(local_model_list.keys()))
        )
        for k in all_models:
            param_str.add_row(
                k,
                str(k in remote_model_list),
                str(k in local_model_list),
                str(remote_model_list.get(k) == local_model_list.get(k)),
            )

        print(param_str)
        local_model_list.clear()
        local_model_list.update(remote_model_list)


threading.Thread(
    target=get_remote_model_list, args=(models_list, False), daemon=True
).start()


def get_device():
    # check if GPU is available

    if torch.cuda.is_available():
        device = torch.device('cuda:0')
    else:
        device = torch.device('cpu')
        warnings.warn(
            '''
!!!!CUDA is not available. DiscoArt is running on CPU. `create()` will be unbearably slow on CPU!!!!
Please switch to a GPU device. If you are using Google Colab, then free tier would just work.
            '''
        )
    return device


def is_jupyter() -> bool:  # pragma: no cover
    """
    Check if we're running in a Jupyter notebook, using magic command `get_ipython` that only available in Jupyter.

    :return: True if run in a Jupyter notebook else False.
    """
    if 'DISCOART_DISABLE_IPYTHON' in os.environ:
        return False

    try:
        get_ipython  # noqa: F821
    except NameError:
        return False
    shell = get_ipython().__class__.__name__  # noqa: F821
    if shell == 'ZMQInteractiveShell':
        return True  # Jupyter notebook or qtconsole
    elif shell == 'Shell':
        return True  # Google colab
    elif shell == 'TerminalInteractiveShell':
        return False  # Terminal running IPython
    else:
        return False  # Other type (?)


def is_google_colab() -> bool:  # pragma: no cover
    if 'DISCOART_DISABLE_IPYTHON' in os.environ:
        return False

    try:
        get_ipython  # noqa: F821
    except NameError:
        return False
    shell = get_ipython().__class__.__name__  # noqa: F821
    return shell == 'Shell'


def get_ipython_funcs(show_widgets: bool = False):
    class NOP:
        def __call__(self, *args, **kwargs):
            return NOP()

        __getattr__ = __enter__ = __exit__ = __iadd__ = __add__ = __call__

    if is_jupyter():
        from IPython import display as dp1
        from IPython.display import FileLink as fl

        handlers = None

        if show_widgets:
            from ipywidgets import HTML, IntProgress, Textarea, Tab

            pg_bar = IntProgress(
                value=1,
                min=0,
                max=4,
                step=1,
                description=f'n_batches:',
                bar_style='info',  # 'success', 'info', 'warning', 'danger' or ''
                orientation='horizontal',
            )
            html_handle = HTML()

            nondefault_config_handle = HTML()
            all_config_handle = HTML()
            completed_handle = HTML()
            completed_handle.value = '<h4>Completed images will be displayed below</h4>'
            code_snippet_handle = Textarea(rows=20)
            tab = Tab()
            tab.children = [
                html_handle,
                nondefault_config_handle,
                all_config_handle,
                code_snippet_handle,
                completed_handle,
            ]
            for idx, j in enumerate(
                (
                    'Preview',
                    'Non-default config',
                    'Full config',
                    'Code snippet',
                    'Completed',
                )
            ):
                tab.set_title(idx, j)

            handlers = SimpleNamespace(
                preview=html_handle,
                config=nondefault_config_handle,
                all_config=all_config_handle,
                code=code_snippet_handle,
                completed=completed_handle,
                progress=pg_bar,
            )

        def redraw():
            dp1.display(pg_bar, tab)

        return dp1, fl, handlers, redraw
    else:
        return NOP(), NOP(), NOP(), NOP()


if not os.path.exists(cache_dir):
    Path(cache_dir).mkdir(parents=True, exist_ok=True)

logger.debug(f'`.cache` dir is set to: {cache_dir}')


def _wget(url, outputdir):
    logger.debug(f'downloading from {url}...')
    try:
        basename = os.path.basename(url)

        with urllib.request.urlopen(url) as source, open(
            os.path.join(outputdir, basename), 'wb'
        ) as output:
            with tqdm(
                total=int(source.info().get("Content-Length")),
                ncols=80,
                unit='iB',
                unit_scale=True,
            ) as loop:
                while True:
                    buffer = source.read(8192)
                    if not buffer:
                        break

                    output.write(buffer)
                    loop.update(len(buffer))
        logger.debug(f'write to {outputdir}')
    except:
        logger.error(f'failed to download {url}')


def load_clip_models(
    device,
    enabled: List[str],
    clip_models: Dict[str, Any] = {},
    text_clip_on_cpu: bool = False,
):
    logger.debug('loading clip models...')

    if text_clip_on_cpu:
        first_device = torch.device('cpu')
        logger.debug(f'CLIP will be first loaded to CPU')
    else:
        first_device = device

    # load enabled models
    for k in enabled:
        if k not in clip_models:
            if '::' in k and (k.split('::')[-1] != 'openai' or '-quickgelu' in k):
                # use open_clip loader
                k1, k2 = k.split('::')
                logger.debug(f'use open_clip to load {k1}')
                import open_clip

                m = open_clip.create_model_and_transforms(
                    k1,
                    device=first_device,
                    pretrained=k2,
                )[0]
            else:

                k1 = (
                    k.replace('B-32::', 'B/32::')
                    .replace('B-16::', 'B/16::')
                    .replace('L-14::', 'L/14::')
                    .replace('L-14-336::', 'L/14@336px::')
                )

                if '::' in k1:
                    k1 = k1.split('::')[0]

                logger.debug(f'use openai clip to load {k1}')

                import clip

                m = clip.load(k1, device=first_device, jit=False)[0]
            clip_models[k] = m.eval().requires_grad_(False)
            if text_clip_on_cpu:
                # then the first device is CPU, we now load visual arm back to GPU.
                m.visual.to(device)
                logger.debug(f'move {k}.visual to GPU')

    # disable not enabled models to save memory
    for k in list(clip_models.keys()):
        if k not in enabled:
            clip_models.pop(k)

    return clip_models


def _check_sha(path, expected_sha):
    if 'DISCOART_DISABLE_CHECK_MODEL_SHA' in os.environ:
        return True
    else:
        with open(path, 'rb') as f:
            return hashlib.sha256(f.read()).hexdigest() == expected_sha


def _get_model_name(name: str) -> str:
    for k in models_list.keys():
        if k.lower().startswith(name.lower().strip()):
            return k


def download_model(model_name: str):
    if os.path.isfile(model_name):
        logger.debug('use customized local model')
        return

    model_name = _get_model_name(model_name)

    if not model_name:
        raise ValueError(
            f'{model_name} is not supported, must be one of {models_list.keys()}'
        )

    model_filename = os.path.basename(models_list[model_name]['sources'][0])
    model_local_path = os.path.join(cache_dir, model_filename)
    if os.path.exists(model_local_path) and _check_sha(
        model_local_path, models_list[model_name]['sha']
    ):
        logger.debug(f'{model_filename} is already downloaded with correct SHA')
    else:
        for url in models_list[model_name]['sources']:
            _wget(url, cache_dir)
            if _check_sha(model_local_path, models_list[model_name]['sha']):
                logger.debug(f'{model_filename} is downloaded with correct SHA')
                break


def get_diffusion_config(user_args, device=torch.device('cuda:0')) -> Dict[str, Any]:
    diffusion_model = user_args.diffusion_model
    steps = user_args.steps
    diffusion_config = user_args.diffusion_model_config

    from guided_diffusion.script_util import (
        model_and_diffusion_defaults,
    )

    model_config = model_and_diffusion_defaults()

    _diff_model_name = _get_model_name(diffusion_model)
    if _diff_model_name and models_list[_diff_model_name].get('config', None):
        model_config.update(models_list[_diff_model_name]['config'])
    else:
        logger.info(
            '''
        looks like you are using a custom diffusion model, 
        to override default diffusion model config, you can specify `create(diffusion_model_config={...}, ...)` as well,
        '''
        )
        model_config.update(
            {
                'attention_resolutions': '16',
                'class_cond': False,
                'diffusion_steps': 1000,
                'rescale_timesteps': True,
                'timestep_respacing': 'ddim100',
                'image_size': 256,
                'learn_sigma': True,
                'noise_schedule': 'linear',
                'num_channels': 128,
                'num_heads': 1,
                'num_res_blocks': 2,
                'use_scale_shift_norm': False,
            }
        )
    timestep_respacing = f'ddim{steps}'
    diffusion_steps = (1000 // steps) * steps if steps < 1000 else steps

    model_config.update(
        {
            'use_fp16': device.type != 'cpu',
            'timestep_respacing': timestep_respacing,
            'diffusion_steps': diffusion_steps,
            'use_checkpoint': True,
        }
    )

    if diffusion_config and isinstance(diffusion_config, dict):
        model_config.update(diffusion_config)

    return model_config


def load_secondary_model(user_args, device=torch.device('cuda:0')):
    if not user_args.use_secondary_model:
        return
    download_model('secondary')

    from discoart.nn.sec_diff import SecondaryDiffusionImageNet2

    secondary_model = SecondaryDiffusionImageNet2()
    secondary_model.load_state_dict(
        torch.load(
            os.path.join(cache_dir, 'secondary_model_imagenet_2.pth'),
            map_location='cpu',
        )
    )
    secondary_model.eval().requires_grad_(False).to(device)
    return secondary_model


def load_diffusion_model(user_args, device):
    diffusion_model = user_args.diffusion_model

    _diff_model_name = _get_model_name(diffusion_model)
    if _diff_model_name:
        rec_size = models_list[_diff_model_name].get('recommended_size', None)
        if rec_size and user_args.width_height != rec_size:
            logger.warning(
                f'{diffusion_model} is recommended to have width_height {rec_size}, but you are using {user_args.width_height}. This may lead to suboptimal results.'
            )

    download_model(diffusion_model)

    model_config = get_diffusion_config(user_args, device=device)

    logger.debug('loading diffusion model...')
    from guided_diffusion.script_util import (
        create_model_and_diffusion,
    )

    model, diffusion = create_model_and_diffusion(**model_config)
    if os.path.isfile(diffusion_model):
        logger.debug(f'loading customized diffusion model from {diffusion_model}')
        _model_path = diffusion_model
    elif _diff_model_name:
        model_filename = os.path.basename(models_list[_diff_model_name]['sources'][0])
        _model_path = os.path.join(cache_dir, model_filename)
    model.load_state_dict(torch.load(_model_path, map_location='cpu'), strict=False)
    model.requires_grad_(False).eval().to(device)

    for name, param in model.named_parameters():
        if 'qkv' in name or 'norm' in name or 'proj' in name:
            param.requires_grad_()
    if model_config['use_fp16']:
        model.convert_to_fp16()

    return model, diffusion


class PromptParser(SimpleTokenizer):
    def __init__(self, on_misspelled_token: str, **kwargs):
        super().__init__(**kwargs)
        self.spell = SpellChecker()
        from . import __resources_path__

        with open(
            os.path.join(__resources_path__, 'vocab.txt'), encoding='utf-8'
        ) as fp:
            self.spell.word_frequency.load_words(
                line.strip() for line in fp if len(line.strip()) > 1
            )
        self.on_misspelled_token = on_misspelled_token

    @staticmethod
    def _split_weight(prompt):
        if ':' in prompt:
            vals = prompt.rsplit(':', 1)
        else:
            vals = [prompt, 1]
        return vals[0], float(vals[1])

    def parse(self, text: str, on_misspelled_token=None) -> Tuple[str, float]:
        text, weight = self._split_weight(text)
        text = whitespace_clean(basic_clean(text)).lower()
        all_tokens = []
        for token in re.findall(self.pat, text):
            token = ''.join(self.byte_encoder[b] for b in token.encode('utf-8'))
            all_tokens.append(token)
        unknowns = [
            v
            for v in self.spell.unknown(all_tokens)
            if len(v) > 2 and self.spell.correction(v) != v
        ]
        if unknowns:
            pairs = []
            for v in unknowns:
                vc = self.spell.correction(v)
                pairs.append((v, vc))
                on_misspelled_token = on_misspelled_token or self.on_misspelled_token
                if on_misspelled_token == 'correct':
                    for idx, ov in enumerate(all_tokens):
                        if ov == v:
                            all_tokens[idx] = vc

            if pairs:
                warning_str = '\n'.join(
                    f'Misspelled `{v}`, do you mean `{vc}`?' for v, vc in pairs
                )
                if on_misspelled_token == 'raise':
                    raise ValueError(warning_str)
                elif on_misspelled_token == 'correct':
                    logger.warning(
                        'auto-corrected the following tokens:\n' + warning_str
                    )
                else:
                    logger.warning(
                        'Found misspelled tokens in the prompt:\n' + warning_str
                    )

        logger.debug(f'prompt: {all_tokens}, weight: {weight}')
        return ' '.join(all_tokens), weight


def free_memory():
    gc.collect()
    torch.cuda.empty_cache()


def show_result_summary(_da, _name, _args):
    from .config import print_args_table

    _dp1, _fl = get_ipython_funcs()[:2]

    _dp1.clear_output(wait=True)

    imcomplete_str = ''

    fully_done = sum(bool(j) for j in _da[:, 'tags___status__completed'])
    num_expected = _args.n_batches * _args.batch_size
    if _da and fully_done < num_expected:
        imcomplete_str = f'''
# ⚠️ Incomplete result ({fully_done}/{num_expected})

Your `n_batches={_args.n_batches}, batch_size={_args.batch_size}` so supposedly {num_expected} images will be generated, 
but only {fully_done} images were fully completed. This may due to the following reasons:
- You cancel the process before it finishes;
- (On Google Colab) your GPU session is expired;

To avoid this, you can set `n_batches` to a smaller number in `create()`, say `create(n_batches=1)`.
'''

    from rich.markdown import Markdown

    md = Markdown(
        f'''
{imcomplete_str}

# 👀 Result preview

This preview is **NOT** in HD. Do **NOT** use it for your final artworks.

To save the full-size images, please check out the instruction in the next section.
    ''',
        code_theme='igor',
    )
    _dp1.display(md)

    if _da and _da[0].uri:
        _da.plot_image_sprites(skip_empty=True, show_index=True, keep_aspect_ratio=True)

    print_args_table(vars(_args))

    persist_file = _fl(
        os.path.join(get_output_dir(_name), 'da.protobuf.lz4'),
        result_html_prefix=f'▶ Download the local backup (in case cloud storage failed): ',
    )

    md = Markdown(
        f'''


# 🖼️ Save images

Final results and intermediate results are created, i.e.
```text
./{_name}/[i]-done.png
./{_name}/[i]-step-[i].png
./{_name}/[i]-progress.gif
./{_name}/[i]-progress.png
./{_name}/da.protobuf.lz4
```

where:

- `i-*` is up to the value of `n_batches`.
- `*-done-*` is the final image on done.
- `*-step-*` is the intermediate image at certain step.
- `*-progress.png` is the sprite image of all intermediate results so far.
- `*-progress.gif` is the animated gif of all intermediate results so far.
- `da.protobuf.lz4` is the LZ4 compressed Protobuf file of all intermediate results of all `n_batches`.

# 💾 Save & load the batch        

Results are stored in a [DocumentArray](https://docarray.jina.ai/fundamentals/documentarray/) available both local and cloud.


You may also download the file manually and load it from local disk:

```python
da = DocumentArray.load_binary('{get_output_dir(_name)}/da.protobuf.lz4')
```

You can simply pull it from any machine:

```python
# pip install docarray[common]
from docarray import DocumentArray

da = DocumentArray.pull('{_name}')
```

More usage such as plotting, post-analysis can be found in the [README](https://github.com/jina-ai/discoart).
            ''',
        code_theme='igor',
    )
    if is_google_colab():
        _dp1.display(md)
    else:
        _dp1.display(persist_file, md)


def list_diffusion_models():
    get_remote_model_list(models_list, force_print=True)


def _version_check(package: str = None, github_repo: str = None):
    try:
        if not package:
            package = vars(sys.modules[__name__])['__package__']
        if not github_repo:
            github_repo = package

        cur_ver = Version(pkg_resources.get_distribution(package).version)
        req = Request(
            f'https://pypi.python.org/pypi/{package}/json',
            headers={'User-Agent': 'Mozilla/5.0'},
        )
        with urlopen(
            req, timeout=1
        ) as resp:  # 'with' is important to close the resource after use
            j = json.load(resp)
            releases = j.get('releases', {})
            latest_release_ver = max(
                Version(v) for v in releases.keys() if '.dev' not in v
            )
            if cur_ver < latest_release_ver:
                from rich import print
                from rich.panel import Panel

                print(
                    Panel(
                        f'You are using [b]{package} {cur_ver}[/b], but [bold green]{latest_release_ver}[/] is available. '
                        f'You may upgrade it via [b]pip install -U {package}[/b]. [link=https://github.com/jina-ai/{github_repo}/releases]Read Changelog here[/link].',
                        title=':new: New version available!',
                        width=50,
                    )
                )
    except Exception as ex:
        # no network, too slow, PyPi is down
        logger.error(f'can not fetch the lastest version number: {ex}')


threading.Thread(target=_version_check, args=(__package__, 'discoart')).start()
_MAX_DIFFUSION_STEPS = 1000


def _is_valid_schedule_str(val) -> bool:
    r = re.match(r'(False\b|True\b|[\(\)\[\]0-9\, \.\*\+\-])+', val)
    if r and r.group(0) == val:
        return True
    return False


def _eval_scheduling_str(val) -> List[float]:
    if isinstance(val, str):
        if _is_valid_schedule_str(val):
            val = eval(val)
        else:
            raise ValueError(
                f'invalid scheduling string: {val}, it contains unsafe code'
            )

    if isinstance(val, (int, float, bool)):
        val = [val] * _MAX_DIFFUSION_STEPS
    elif isinstance(val, (list, tuple)):
        if len(val) != _MAX_DIFFUSION_STEPS:
            raise ValueError(
                f'invalid scheduling string: {val} the schedule steps should be exactly {_MAX_DIFFUSION_STEPS}'
            )
    else:
        raise ValueError(f'unsupported scheduling type: {val}: {type(val)}')
    return val


def _get_current_schedule(schedule_table: Dict, t: int) -> 'SimpleNamespace':
    return SimpleNamespace(**{k: schedule_table[k][t] for k in schedule_table.keys()})


def _get_schedule_table(args) -> Dict:
    return {
        k: _eval_scheduling_str(getattr(args, k))
        for k in (
            'cut_overview',
            'cut_innercut',
            'cut_icgray_p',
            'cut_ic_pow',
            'use_secondary_model',
            'cutn_batches',
            'clip_guidance_scale',
            'tv_scale',
            'range_scale',
            'sat_scale',
            'init_scale',
            'clamp_grad',
            'clamp_max',
        )
    }


def get_output_dir(name_da):
    output_dir = os.path.join(os.environ.get('DISCOART_OUTPUT_DIR', './'), name_da)
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    return output_dir



import copy
import os.path
import tempfile
import threading
from typing import Callable, Optional

import clip
import lpips
import torch
import torchvision.transforms.functional as TF
import wandb
from docarray import DocumentArray, Document
from torch.nn.functional import normalize as normalize_fn

from .config import save_config_svg, export_python
from .helper import (
    logger,
    get_ipython_funcs,
    free_memory,
    _MAX_DIFFUSION_STEPS,
    _eval_scheduling_str,
    _get_current_schedule,
    _get_schedule_table,
    get_output_dir,
    is_jupyter,
)
from .nn.helper import set_seed, detach_gpu
from .nn.losses import spherical_dist_loss, tv_loss, range_loss
from .nn.make_cutouts import MakeCutouts
from .nn.sec_diff import alpha_sigma_to_t
from .nn.transform import symmetry_transformation_fn, inv_normalize
from .persist import _sample_thread, _persist_thread, _save_progress_thread
from .prompt import PromptPlanner


def do_run(
    args, models, device, events, image_callback: Optional[Callable[[str], None]] = None
) -> 'DocumentArray':
    skip_event, stop_event = events

    _is_jupyter = is_jupyter()

    output_dir = get_output_dir(args.name_docarray)

    logger.info('preparing models...')

    model, diffusion, clip_models, secondary_model = models
    lpips_model = lpips.LPIPS(net='vgg').to(device)

    side_x, side_y = ((args.width_height[j] // 64) * 64 for j in (0, 1))

    schedule_table = _get_schedule_table(args)

    from .nn.perlin_noises import regen_perlin

    skip_steps = args.skip_steps

    loss_values = []

    model_stats = []

    _dp1, _, _handlers, _redraw_fn = get_ipython_funcs(show_widgets=True)
    _dp1.clear_output(wait=True)

    prompts = PromptPlanner(args)

    text_device = torch.device('cpu') if args.text_clip_on_cpu else device

    for model_name, clip_model in clip_models.items():

        # when using SLIP Base model the dimensions need to be hard coded to avoid AttributeError: 'VisionTransformer' object has no attribute 'input_resolution'
        try:
            try:
                input_resolution = clip_model.visual.input_resolution
            except:
                input_resolution = clip_model.visual.image_size
            logger.debug(f'input_resolution of {model_name}: {input_resolution}')
        except:
            input_resolution = 224
            logger.debug(
                f'fail to find input_resolution for {model_name}, fall back to {input_resolution}'
            )

        schedules = [True] * _MAX_DIFFUSION_STEPS
        if args.clip_models_schedules and model_name in args.clip_models_schedules:
            schedules = _eval_scheduling_str(args.clip_models_schedules[model_name])

        clip_model_stats = {
            'model_name': model_name,
            'clip_model': clip_model,
            'prompt_embeds': [],
            'schedules': schedules,
            'input_resolution': input_resolution,
        }

        for _p in prompts:
            txt = clip_model.encode_text(
                clip.tokenize(
                    _p.tokenized, truncate=args.truncate_overlength_prompt
                ).to(text_device)
            )

            clip_model_stats['prompt_embeds'].append(txt)

        clip_model_stats['prompt_embeds'] = torch.cat(
            clip_model_stats['prompt_embeds']
        ).to(device)

        model_stats.append(clip_model_stats)

    init = None

    set_seed(args.seed)
    if args.init_image:
        d = Document(uri=args.init_image).load_uri_to_image_tensor(side_x, side_y)
        init = (
            TF.to_tensor(d.tensor)
            .to(device)
            .unsqueeze(0)
            .mul(2)
            .sub(1)
            .expand(args.batch_size, -1, -1, -1)
        )

    cur_t = None

    def cond_fn(x, t, **kwargs):

        t_int = (
            int(t[0].item()) + 1
        )  # errors on last step without +1, need to find source

        num_step = _MAX_DIFFUSION_STEPS - t_int
        scheduler = _get_current_schedule(schedule_table, num_step)
        is_cuts_visualized = False

        with torch.enable_grad():

            x = x.detach().requires_grad_()
            if scheduler.use_secondary_model:
                alpha = torch.tensor(
                    diffusion.sqrt_alphas_cumprod[cur_t],
                    device=device,
                    dtype=torch.float32,
                )
                sigma = torch.tensor(
                    diffusion.sqrt_one_minus_alphas_cumprod[cur_t],
                    device=device,
                    dtype=torch.float32,
                )
                cosine_t = alpha_sigma_to_t(alpha, sigma)
                out = secondary_model(x, cosine_t[None].repeat([x.shape[0]])).pred
            else:
                my_t = torch.ones([x.shape[0]], device=device, dtype=torch.long) * cur_t
                out = diffusion.p_mean_variance(model, x, my_t, clip_denoised=False)[
                    'pred_xstart'
                ]

            fac = diffusion.sqrt_one_minus_alphas_cumprod[cur_t]
            x_in = out * fac + x * (1 - fac)

            if scheduler.tv_scale:
                tv_losses = tv_loss(x_in).sum() * scheduler.tv_scale
            else:
                tv_losses = 0

            if scheduler.range_scale:
                range_losses = range_loss(x_in).sum() * scheduler.range_scale
            else:
                range_losses = 0

            if scheduler.sat_scale:
                sat_losses = (
                    torch.abs(x_in - x_in.clamp(min=-1, max=1)).mean().sum()
                    * scheduler.sat_scale
                )
            else:
                sat_losses = 0

            if init is not None and scheduler.init_scale:
                init_losses = lpips_model(x_in, init).sum() * scheduler.init_scale
            else:
                init_losses = 0

            loss = tv_losses + range_losses + sat_losses + init_losses

            if loss != 0:
                x_in_grad = torch.autograd.grad(loss, x_in)[0]
            else:
                x_in_grad = 0

            cut_losses = 0

            for model_stat in model_stats:

                if not model_stat['schedules'][num_step]:
                    continue

                active_prompt_ids = prompts.get_prompt_ids(
                    model_stat['model_name'], num_step
                )

                if active_prompt_ids:
                    masked_embeds = model_stat['prompt_embeds'][
                        list(active_prompt_ids[0])
                    ]
                    masked_weights = normalize_fn(
                        torch.tensor(
                            active_prompt_ids[1], device=device, dtype=torch.float16
                        ),
                        dim=0,
                    )
                    logger.debug(
                        f'activate prompt ids: {active_prompt_ids} prompt weights: {masked_weights}'
                    )
                else:
                    continue

                cuts = MakeCutouts(
                    model_stat['input_resolution'],
                    Overview=scheduler.cut_overview,
                    InnerCrop=scheduler.cut_innercut,
                    IC_Size_Pow=scheduler.cut_ic_pow,
                    IC_Grey_P=scheduler.cut_icgray_p,
                )

                for _ in range(scheduler.cutn_batches):

                    clip_in = cuts(x_in.add(1).div(2))

                    if args.visualize_cuts and not is_cuts_visualized:
                        _cuts_da = DocumentArray.empty(clip_in.shape[0])
                        _cuts_da.tensors = (
                            (inv_normalize(clip_in) * 255).detach().cpu().numpy()
                        )
                        _cuts_da.plot_image_sprites(
                            os.path.join(output_dir, f'{_nb}-cuts-{num_step}.png'),
                            show_index=True,
                            channel_axis=0,
                        )
                        is_cuts_visualized = True

                    image_embeds = (
                        model_stat['clip_model'].encode_image(clip_in).unsqueeze(1)
                    )

                    dists = spherical_dist_loss(
                        image_embeds,
                        masked_embeds.unsqueeze(0),  # 1, 2, 512
                    )

                    dists = dists.view(
                        [
                            scheduler.cut_overview + scheduler.cut_innercut,
                            x.shape[0],
                            -1,
                        ]
                    )

                    cut_loss = (
                        dists.mul(masked_weights).sum(2).mean(0).sum()
                        * scheduler.clip_guidance_scale
                        / scheduler.cutn_batches
                    )

                    x_in_grad += torch.autograd.grad(cut_loss, x_in)[0]

                    cut_losses += cut_loss.detach().item()

        x_is_NaN = False
        if isinstance(x_in_grad, int) and x_in_grad == 0:
            grad = torch.zeros_like(x)
        elif not torch.isnan(x_in_grad).any():
            grad = -torch.autograd.grad(x_in, x, x_in_grad)[0]
        else:
            x_is_NaN = True
            grad = torch.zeros_like(x)
            logger.warning(
                f'NaN detected in grad at the diffusion inner-step {num_step}, no panic. '
                f'However, if this message continues to show up *in a row*, '
                f'then your generation is ill-conditioned and image will not updated, further steps are unnecessary.'
            )

        r_grad = grad
        if scheduler.clamp_grad and not x_is_NaN:
            magnitude = r_grad.square().mean().sqrt()
            r_grad = (
                grad * magnitude.clamp(max=scheduler.clamp_max) / magnitude
            )  # min=-0.02, min=-clamp_max,

        traced_info = {
            'losses/total': detach_gpu(loss) + cut_losses,
            'losses/tv': detach_gpu(tv_losses),
            'losses/range': detach_gpu(range_losses),
            'losses/sat': detach_gpu(sat_losses),
            'losses/init': detach_gpu(init_losses),
            'losses/cuts': detach_gpu(cut_losses),
        }

        traced_info.update(
            {
                f'scheduler/{k}': int(v) if isinstance(v, bool) else v
                for k, v in vars(scheduler).items()
            }
        )

        try:
            traced_info['gradients'] = wandb.Histogram(r_grad.detach().cpu().numpy())
        except ValueError:
            # avoid nan gradients
            pass

        wandb.log(traced_info)
        loss_values.append(traced_info['losses/total'])

        return r_grad

    if args.diffusion_sampling_mode == 'ddim':
        sample_fn = diffusion.ddim_sample_loop_progressive
    else:
        sample_fn = diffusion.plms_sample_loop_progressive

    is_busy_evs = [threading.Event() for _ in range(3)]

    da_batches = DocumentArray()

    org_seed = args.seed

    if os.environ.get('WANDB_MODE', 'disabled') == 'disabled':
        logger.info(
            '''
W&B dashboard is disabled. To enable the online dashboard for tracking losses, gradients, 
scheduling tracking, please set `WANDB_MODE=online` before running/importing DiscoArt. e.g.

    import os
    os.environ['WANDB_MODE'] = 'online'

    from discoart import create
    create(...)
'''
        )

    for _nb in range(args.n_batches):
        logger.info(
            f'creating artworks `{args.name_docarray}` ({_nb}/{args.n_batches})...'
        )

        # set seed for each image in the batch
        new_seed = org_seed + _nb
        set_seed(new_seed)
        args.seed = new_seed
        if _is_jupyter:
            redraw_widget(
                _handlers,
                _redraw_fn,
                args,
                _nb,
            )
        free_memory()

        _da = DocumentArray(
            [Document(tags=copy.deepcopy(vars(args))) for _ in range(args.batch_size)]
        )
        _da_gif = DocumentArray([Document() for _ in range(args.batch_size)])
        da_batches.extend(_da)

        cur_t = diffusion.num_timesteps - skip_steps - 1

        if args.perlin_init:
            init = regen_perlin(
                args.perlin_mode, side_y, side_x, device, args.batch_size
            )

        if args.diffusion_sampling_mode == 'ddim':
            samples = sample_fn(
                model,
                (args.batch_size, 3, side_y, side_x),
                clip_denoised=args.clip_denoised,
                model_kwargs={},
                cond_fn=cond_fn,
                progress='DISCOART_DISABLE_TQDM' not in os.environ,
                skip_timesteps=skip_steps,
                init_image=init,
                randomize_class=args.randomize_class,
                eta=args.eta,
                transformation_fn=lambda x: symmetry_transformation_fn(
                    x, args.use_horizontal_symmetry, args.use_vertical_symmetry
                ),
                transformation_percent=args.transformation_percent,
            )
        else:
            samples = sample_fn(
                model,
                (args.batch_size, 3, side_y, side_x),
                clip_denoised=args.clip_denoised,
                model_kwargs={},
                cond_fn=cond_fn,
                progress='DISCOART_DISABLE_TQDM' not in os.environ,
                skip_timesteps=skip_steps,
                init_image=init,
                randomize_class=args.randomize_class,
                order=2,
            )

        threads = []

        with wandb.init(
            project=args.name_docarray,
            config=vars(args),
            anonymous='must',
            reinit=True,
            mode=os.environ.get('WANDB_MODE', 'disabled'),
        ):
            for j, sample in enumerate(samples):
                if skip_event.is_set() or stop_event.is_set():
                    logger.debug('skip_event/stop_event is set, skipping this run')
                    skip_event.clear()
                    break

                cur_t -= 1

                is_save_step = args.save_rate > 0 and j % args.save_rate == 0
                is_complete = cur_t == -1
                is_display_step = args.display_rate > 0 and j % args.display_rate == 0

                threads.append(
                    _sample_thread(
                        sample,
                        _nb,
                        cur_t,
                        _da,
                        _da_gif,
                        _handlers,
                        j,
                        loss_values,
                        output_dir,
                        is_busy_evs[0],
                        is_save_step or is_complete,
                        args.gif_fps > 0,
                        args.image_output,
                        is_display_step,
                        image_callback,
                    )
                )

                if is_complete or is_save_step:
                    if args.image_output:
                        threads.append(
                            _save_progress_thread(
                                _da,
                                _da_gif,
                                _nb,
                                output_dir,
                                args.gif_fps,
                                args.gif_size_ratio,
                            )
                        )

                    threads.extend(
                        _persist_thread(
                            da_batches,
                            args.name_docarray,
                            is_busy_evs[1:],
                            is_busy_evs[0],
                            is_completed=is_complete,
                        )
                    )

        for t in threads:
            t.join()
        _dp1.clear_output(wait=True)

        if stop_event.is_set():
            logger.debug('stop_event is set, skipping the while `n_batches`')
            stop_event.clear()
            break

    logger.info(f'done! {args.name_docarray}')

    return da_batches


def redraw_widget(_handlers, _redraw_fn, args, _nb):
    _handlers.progress.max = args.n_batches
    _handlers.progress.value = _nb + 1
    _handlers.progress.description = f'Baking {_nb + 1}/{args.n_batches}: '

    svg_name = f'{os.path.join(tempfile.gettempdir(), args.name_docarray)}.svg'
    save_config_svg(args, svg_name, only_non_default=True)
    d = Document(uri=svg_name).convert_uri_to_datauri()
    _handlers.config.value = f'<img src="{d.uri}" alt="non-default config">'

    save_config_svg(args, svg_name)
    d = Document(uri=svg_name).convert_uri_to_datauri()
    _handlers.all_config.value = f'<img src="{d.uri}" alt="all config">'

    _handlers.code.value = export_python(args)
    _redraw_fn()



import copy
import os
import random
import uuid
from types import SimpleNamespace
from typing import Dict, Union, Optional, Tuple

import yaml
from docarray import DocumentArray, Document
from yaml import Loader

from . import __resources_path__, __version__

with open(
    os.environ.get(
        'DISCOART_DEFAULT_PARAMETERS_YAML',
        os.path.join(__resources_path__, 'default.yml'),
    )
) as ymlfile:
    default_args = yaml.load(ymlfile, Loader=Loader)

with open(
    os.environ.get(
        'DISCOART_CUT_SCHEDULES_YAML',
        os.path.join(__resources_path__, 'cut-schedules.yml'),
    )
) as ymlfile:
    cut_schedules = yaml.load(ymlfile, Loader=Loader)

_legacy_args = {'clip_sequential_evaluation', 'fuzzy_prompt', 'skip_augs'}


def load_config(
    user_config: Union[Dict, str],
) -> Dict:
    if not isinstance(user_config, dict):
        with open(user_config) as f:
            user_config = yaml.load(f, Loader=Loader)

    cfg = copy.deepcopy(default_args)

    for k in list(user_config.keys()):
        if k not in cfg and not k.startswith('_') and k not in _legacy_args:
            raise AttributeError(f'unknown argument `{k}`, misspelled?')
        if k in _legacy_args or k.startswith('_'):
            # remove private arguments in tags
            user_config.pop(k)

    if user_config:
        if user_config.get('cut_schedules_group', None):
            cfg.update(cut_schedules[user_config['cut_schedules_group']])

        cfg.update(**user_config)

    int_keys = {k for k, v in default_args.items() if isinstance(v, int)}
    int_keys.add('seed')

    for k, v in cfg.items():
        if k in int_keys and v is not None and not isinstance(v, (int, str)):
            cfg[k] = int(v)
        if k == 'width_height':
            cfg[k] = [int(vv) for vv in v]

    cfg.update(
        **{
            'seed': int(cfg['seed'] or random.randint(0, 2**32)),
        }
    )

    _id = uuid.uuid1().hex
    if cfg['batch_name']:
        da_name = f'{__package__}-{cfg["batch_name"]}-{_id}'
    else:
        da_name = f'{__package__}-{_id}'

    if not cfg.get('name_docarray', None):
        cfg['name_docarray'] = da_name

    cfg['name_docarray'] = cfg['name_docarray'].format(**cfg)
    return cfg


def show_config(
    docs: Union['Document', 'Document', Dict, str, SimpleNamespace],
    only_non_default: bool = True,
):
    cfg = _extract_config_from_docs(docs)
    print_args_table(cfg, only_non_default=only_non_default)


def save_config_svg(
    docs: Union['Document', 'Document', Dict, str, SimpleNamespace],
    output: Optional[str] = None,
    **kwargs,
) -> None:
    """
    Save the config as SVG.
    :param docs: a DocumentArray or Document or a Document.tags dict
    :param output: the filename to store the SVG, if not given, it will be saved as `{name_docarray}.svg`
    :return:
    """
    cfg = _extract_config_from_docs(docs)

    from rich.console import Console
    from rich.terminal_theme import MONOKAI
    from tempfile import NamedTemporaryFile

    with NamedTemporaryFile(mode='wt') as fp:
        console = Console(
            record=True, file=fp, force_jupyter=False, force_interactive=False
        )
        print_args_table(cfg, console, **kwargs)
        console.save_svg(
            output or f'{cfg["name_docarray"]}.svg',
            theme=MONOKAI,
            title=f'DiscoArt {__version__}',
        )


def print_args_table(
    cfg,
    console=None,
    only_non_default: bool = False,
    console_print: bool = True,
    table_title: Optional[str] = None,
):
    from rich.table import Table
    from rich import box
    from rich.console import Console

    if console is None:
        console = Console()

    param_str = Table(
        title=table_title or cfg['name_docarray'],
        caption=f'showing only non-default args'
        if only_non_default
        else 'showing all args ([b]bold *[/] args are non-default)',
        box=box.ROUNDED,
        highlight=True,
        title_justify='center',
    )
    param_str.add_column('Argument', justify='right')
    param_str.add_column('Value', justify='left')

    for k, v in sorted(cfg.items()):
        if k.startswith('_'):
            continue
        value = str(v)
        _non_default = False
        if not default_args.get(k, None) == v:
            if not only_non_default:
                k = f'[b]{k}*[/]'
            _non_default = True

        if not only_non_default or _non_default:
            param_str.add_row(k, value)

    if console_print:
        console.print(param_str)
    return param_str


def cheatsheet():
    from . import __resources_path__

    from rich.table import Table
    from rich import box
    from rich.console import Console

    console = Console()

    with open(os.path.join(__resources_path__, 'docstrings.yml')) as ymlfile:
        docs = yaml.load(ymlfile, Loader=Loader)

    param_tab = Table(
        title=f'Cheatsheet for all supported parameters',
        box=box.ROUNDED,
        highlight=True,
        show_lines=True,
        title_justify='center',
    )
    param_tab.add_column('Argument', justify='right')
    param_tab.add_column('Default', justify='left', max_width=10, overflow='fold')
    param_tab.add_column('Description', justify='left')

    for k, v in sorted(default_args.items()):
        value = str(v)
        if k in docs:
            d_string = docs[k]
            param_tab.add_row(
                str(k),
                value,
                d_string.replace(
                    '[DiscoArt]', '[bold black on yellow]:new: DiscoArt[/]'
                ),
            )

    console.print(param_tab)


def save_config(
    docs: Union['Document', 'Document', Dict, str, SimpleNamespace], output: str
) -> None:
    cfg = _extract_config_from_docs(docs)
    with open(output, 'w') as f:
        yaml.dump(cfg, f)


def _extract_config_from_docs(docs):
    cfg = None

    if isinstance(docs, DocumentArray):
        cfg = docs[0].tags
    elif isinstance(docs, Document):
        cfg = docs.tags
    elif isinstance(docs, dict):
        cfg = docs
    elif isinstance(docs, SimpleNamespace):
        cfg = vars(docs)
    elif isinstance(docs, str):
        cfg = DocumentArray.pull(docs)[0].tags

    return load_config(cfg)


def export_python(
    docs: Union['Document', 'Document', Dict, str, SimpleNamespace],
    ignored_args: Tuple[str] = ('name_docarray',),
) -> str:
    cfg = _extract_config_from_docs(docs)
    non_defaults = {}
    for k, v in cfg.items():
        if k.startswith('_') or k in ignored_args:
            continue

        if not default_args.get(k, None) == v:
            non_defaults[k] = v

    kwargs_string = ',\n    '.join(
        f'{k}=\'{v}\'' if isinstance(v, str) else f'{k}={v}'
        for k, v in non_defaults.items()
    )

    return f'''
#!pip install {__package__}=={__version__}

from discoart import create

da = create(
    {kwargs_string}
)'''



import numpy as np
import torch
import random


def set_seed(seed: int) -> None:
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True


def detach_gpu(val):
    if isinstance(val, (int, float)):
        return val
    else:
        return val.detach().cpu().item()



import torch
from resize_right import resize
from torch import nn
from torch.nn import functional as F
from torchvision import transforms as T
from torchvision.transforms import functional as TF


class MakeCutouts(nn.Module):
    def __init__(
        self,
        cut_size,
        Overview=4,
        InnerCrop=0,
        IC_Size_Pow=0.5,
        IC_Grey_P=0.2,
    ):
        super().__init__()
        self.cut_size = cut_size
        self.Overview = Overview
        self.InnerCrop = InnerCrop
        self.IC_Size_Pow = IC_Size_Pow
        self.IC_Grey_P = IC_Grey_P
        self.augment = T.Compose(
            [
                T.RandomHorizontalFlip(p=0.5),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                T.RandomAffine(
                    degrees=10,
                    translate=(0.05, 0.05),
                    interpolation=T.InterpolationMode.BILINEAR,
                ),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                T.RandomGrayscale(p=0.1),
                T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
                T.ColorJitter(brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1),
                T.Normalize(
                    mean=[0.48145466, 0.4578275, 0.40821073],
                    std=[0.26862954, 0.26130258, 0.27577711],
                ),
            ]
        )

    def forward(self, input):
        return torch.cat([self.augment(c) for c in self._cut_generator(input)])

    def _cut_generator(self, input):
        gray = T.Grayscale(3)
        sideY, sideX = input.shape[2:4]
        max_size = min(sideX, sideY)
        min_size = min(sideX, sideY, self.cut_size)
        output_shape = [input.shape[0], 3, self.cut_size, self.cut_size]
        pad_input = F.pad(
            input,
            (
                (sideY - max_size) // 2,
                (sideY - max_size) // 2,
                (sideX - max_size) // 2,
                (sideX - max_size) // 2,
            ),
        )

        cutout = resize(pad_input, out_shape=output_shape)
        for j in range(self.Overview):
            if j == 1:
                yield gray(cutout)
            elif j == 2:
                yield TF.hflip(cutout)
            elif j == 3:
                yield gray(TF.hflip(cutout))
            else:
                yield cutout

        for i in range(self.InnerCrop):
            size = int(
                torch.rand([]) ** self.IC_Size_Pow * (max_size - min_size) + min_size
            )
            offsetx = torch.randint(0, sideX - size + 1, ())
            offsety = torch.randint(0, sideY - size + 1, ())
            cutout = input[:, :, offsety : offsety + size, offsetx : offsetx + size]
            if i <= int(self.IC_Grey_P * self.InnerCrop):
                cutout = gray(cutout)
            yield resize(cutout, out_shape=output_shape)



import torch
import torchvision.transforms as T

inv_normalize = T.Normalize(
    mean=[-0.48145466 / 0.26862954, -0.4578275 / 0.26130258, -0.40821073 / 0.27577711],
    std=[1 / 0.26862954, 1 / 0.26130258, 1 / 0.27577711],
)


def symmetry_transformation_fn(x, use_horizontal_symmetry, use_vertical_symmetry):
    if use_horizontal_symmetry:
        [n, c, h, w] = x.size()
        x = torch.concat(
            (x[:, :, :, : w // 2], torch.flip(x[:, :, :, : w // 2], [-1])), -1
        )
    if use_vertical_symmetry:
        [n, c, h, w] = x.size()
        x = torch.concat(
            (x[:, :, : h // 2, :], torch.flip(x[:, :, : h // 2, :], [-2])), -2
        )
    return x
